% processes.pl - Semantics of process algebra specifications.
%
% (c) 2003 Martin Pitt <martin@piware.de>
%
% This software can be used freely under the terms and conditions of
% the GNU Lesser General Public License (LGPL). See
% http://www.gnu.org/copyleft/lgpl.html for details.
%
% -------------------------------------------------------------------
% Notational translation:
%
% Transition	    PROLOG
% ----------	    ------
% out := f(x,y)	    out :== f(x,y)	(functional assignment)
% param = c	    param == c		(parameter setting)
% no translation: value assignment (:=), input (=), update(), send()
%
% Process	    PROLOG
% -------	    ------
% 0		    0
% a1 -> a2 -> P     [a1, a2, p]
% stub.ch -> P	    stub^ch -> P	 (qualification)
% <P1> | <P2>	    <P1> + <P2>
% [prop1] P	    [prop1,...] * P	 (multiple props allowed)	
% K := <P>	    processdef( k, <P> )
%
% If process definitions employ variables, these can be instantiated
% in the processdef/3 form in the 3rd parameter:
% K(x) := <P(y)>    processdef( k(X), <P(Y)>, Y is f(X) )
%
% Networks are defined by net( Name, Stubs, Bind, Init ) with
%    Name:  network name (atom)
%    Stubs: [stubname:process, ...] list of stub instantiations
%    Bind:  [stub^out/stub/in, ...] list of channel bindings
%    Init:  [delta1, delta2, ...]   list of initial actions 
%				    (for parameter setting)
%
% -------------------------------------------------------------------
% Imported clauses:
%   processdef( Constant, Definition )   (see above)
%   net( Constant, Stubs, Bind, Init )   (see above)
% ---
% Exported clauses:
%   next( Process, Transition, Process' )
%   has_property( Process, Property )
%   expand_proc_const( Constant, Definition ) (Constant may be
%                                              preceeded by properties)
%   proc_equiv( Process, Process )
%   poss_trans_but( Process, [Delta,...], [Delta,...] )
% -------------------------------------------------------------------

% operators for transitions
:- op( 300, xfx, := ). % output value assignment
:- op( 300, xfx, :== ). % output functional assignment

:- discontiguous( processdef/2, processdef/3, net/4 ).

% provide dummy clauses in case none are defined externally
processdef( _, _, _ ) :- fail.
net( _, _, _, _ ) :- fail.

% transition classification
action(A) :- A = _:=_; A = _:==_; A = send(_).
event(E) :- E =.. [=,_,_]; E =.. [==,_,_]; E = update(_). 
waiting(W) :- W = tick; W = tick(_).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Process semantics - define the relation next( P, Delta, P' ).
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% drop brackets around a single process name
next( [P], A, P1 ) :- !, next( P, A, P1 ).

% basic operators
next( 0, _, _ ) :- !, fail.
next( [A,P], A, P ) :- (action(A); event(A); A = tick), !.
next( [A|P], A, P ) :- action(A); event(A); A = tick.
next( PA + PB, A, P2 ) :- next( PA, A, P2 ); next( PB, A, P2 ).

% update reacts to any input value
next( [update(I)|P], I=C, P ) :- nonvar(C).

% tick(N) means n ticks
next( [tick(1)|P], tick, P ).
next( [tick(N)|P], tick, [tick(N1)|P] ) :- N > 1, N1 is N - 1.

% atomic properties: keep if no action (tick), drop otherwise
next( Prop * P, tick, Prop * P1 ) :- next( P, tick, P1 ).
next( _ * P, A, P1 ) :- ( action(A); event(A) ), next( P, A, P1 ).

% substitute definition for constant name
next( K, A, P2 ) :- expand_proc_const( K, P ), next( P, A, P2 ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Network process semantics
%
% This defines the relation next( P, Delta, P' ) for networks of
% processes.
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

next( n(K, []), _, n(K, []) ).

% Actions

next( n(K, Stubs), St^Out:=C, n(K, [St:PSt1|Stubs1] ) ) :-
    getstub( Stubs, St, PSt, StubsR ),
    next( PSt, Out:=C, PSt1 ),
    ( bound( St^Out, _^_, K ) ->
	bindev( K, StubsR, St^Out:=C, Stubs1 );
	posstick_net( StubsR, Stubs1 )
    ).

next( n(K, Stubs), St^Out:==F, n(K, [St:PSt1|Stubs1] ) ) :-
    getstub( Stubs, St, PSt, StubsR ),
    next( PSt, Out:==F, PSt1 ),
    ( bound( St^Out, _^_, K ) ->
	bindev( K, StubsR, St^Out:==F, Stubs1 );
	posstick_net( StubsR, Stubs1 )
    ).

next( n(K, Stubs), send(St^Out), n(K, [St:PSt1|Stubs1] ) ) :-
    getstub( Stubs, St, PSt, StubsR ),
    next( PSt, send(Out), PSt1 ),
    ( bound( St^Out, _^_, K ) ->
	bindev( K, StubsR, send(St^Out), Stubs1 );
	posstick_net( StubsR, Stubs1 )
    ).

% Events of bound channels

next( n(K, Stubs), St^In=C, n(K, Stubs1 ) ) :-
    getstub( Stubs, St, PSt, _ ),
    next( PSt, In=C, _ ),
    bound( OS^Out, St^In, K ),
    next( n(K, Stubs), OS^Out:=C, n(K, Stubs1 ) ),
    nonvar(C).

next( n(K, Stubs), update(St^In), n(K, Stubs1 ) ) :-
    getstub( Stubs, St, PSt, _ ),
    next( PSt, update(In), _ ),
    bound( OS^Out, St^In, K ),
    (
	next( n(K, Stubs), OS^Out:=_, n(K, Stubs1 ) );
	next( n(K, Stubs), OS^Out:==_, n(K, Stubs1 ) );
	next( n(K, Stubs), send(OS^Out), n(K, Stubs1 ) )
    ).

% Events of unbound channels

next( n(K, Stubs), St^In=C, n(K, [St:PSt1|Stubs1] ) ) :-
    getstub( Stubs, St, PSt, StubsR ),
    next( PSt, In=C, PSt1 ),
    \+ bound( _^_, St^In, K ),
    posstick_net( StubsR, Stubs1 ),
    nonvar(C).

next( n(K, Stubs), update(St^In), n(K, [St:PSt1|Stubs1] ) ) :-
    getstub( Stubs, St, PSt, StubsR ),
    next( PSt, update(In), PSt1 ),
    \+ bound( _^_, St^In, K ),
    posstick_net( StubsR, Stubs1 ).

next( n(K, Stubs), A, n(K, [St:PSt1|Stubs1] ) ) :-
    nonvar(A), A = (St^Par==C),  % prevent search for parameter changes
    getstub( Stubs, St, PSt, StubsR ),
    next( PSt, Par==C, PSt1 ),
    posstick_net( StubsR, Stubs1 ).

% tick

next( n(K, []), tick, n(K, []) ).

next( n(K, [S:P|Stubs]), tick, n(K, [S:P1|Stubs1]) ) :-
    ( next( P, tick, P1 ) -> true; \+ ( action(A), next(P,A,_) ), P1=P ),
    next( n(K, Stubs), tick, n(K, Stubs1) ).

% helper predicates

posstick_stub( P, P1 ) :- next( P, tick, P2 ) -> P1 = P2; P1 = P.

posstick_net( [], [] ) :- !.
posstick_net( [S:P|Stubs], [S:P1|Stubs1] ) :-
    posstick_stub( P, P1 ),
    posstick_net( Stubs, Stubs1 ).

bindev( Net, St, P, Os^Out:=C, P1 ) :-
    bound( Os^Out, St^In, Net ),
    next( P, In=C, P1 ).

bindev( Net, St, P, send(Os^Out), P1 ) :-
    bound( Os^Out, St^In, Net ),
    next( P, update(In), P1 ).

bindev( Net, St, P, Os^Out:==_, P1 ) :-
    bound( Os^Out, St^In, Net ),
    next( P, update(In), P1 ).

bindev( _, [], _, [] ).
bindev( Net, [S:P|Stubs], A, [S:P1|Stubs1] ) :-
    ( bindev( Net, S, P, A, P2 ) ->
	P1 = P2;
	posstick_stub( P, P1 )
    ), 
    bindev( Net, Stubs, A, Stubs1 ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Atomic property semantics: has_property( Process, Property )
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% qualified property in networks
has_property( n(_,Stubs), St^Prop ) :-
    getstub( Stubs, St, P, _ ),
    has_property( P, Prop ).

has_property( n(_,Stubs), Prop ) :-
    getstub( Stubs, _, P, _ ),
    has_property( P, Prop ).

has_property( [Prop|_] * _, Prop ).
has_property( [_|Props] * P, Prop ) :- has_property( Props * P, Prop ).
has_property( [] * P, Prop ) :- has_property( P, Prop ).

has_property( P1 + P2, Prop ) :- 
    has_property( P1, Prop ); has_property( P2, Prop ).

has_property( K, Prop ) :- 
    expand_proc_const( K, P ), has_property( P, Prop ).

has_property( [P], Prop ) :- has_property( P, Prop ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Helper predicates
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% succeeds iff In is bound to Out in network Net
bound( Out, In, Net ) :- net( Net, _, Bind, _ ), boundlist( Out, Bind, In ).

boundlist( X, [X/Y|_], Y ).
boundlist( X, [_|M], Y ) :- boundlist( X, M, Y ).

% getstub( [s1:P1,...,sn:Pn], S, P, R )
% find S:P in the Stub list, with R being the stub list without s:P
getstub( [S:P|R], S, P, R ).
getstub( [S1:P1|Stubs], S, P, [S1:P1|R] ) :- 
    getstub( Stubs, S, P, R ).

% Constant definition expansion

expand_proc_const( Prop * K, Prop * P ) :- expand_proc_const( K, P ).

expand_proc_const( K, P ) :-
    processdef( K, Def ) -> P = Def;
    processdef( K, Def, Inst ) -> call( Inst ), P = Def;
    net( K, Processes, _, Init ) -> nextlist( n(K,Processes), Init, P ).

% next for a list of actions
nextlist( P, [], P ).
nextlist( P, [A|As], PE ) :- next( P, A, P1 ), nextlist( P1, As, PE ).

% process equivalency
proc_equiv( P, P ) :- !.
proc_equiv( P, [P1] ) :- proc_equiv( P, P1 ).
proc_equiv( [P1], P ) :- proc_equiv( P, P1 ).
proc_equiv( K, P ) :- expand_proc_const( K, P1 ), !, proc_equiv( P1, P ).

proc_equiv( n(K, []), n(K, []) ).
proc_equiv( n(K, [Stub:P|Stubs]), n(K,Stubs1) ) :- 
    getstub( Stubs1, Stub, P1, Stubs1R ),
    proc_equiv( P, P1 ), 
    proc_equiv( n(K, Stubs), n(K, Stubs1R) ).

% Trans <- possible transitions of P, excluding the ones in Excl
poss_trans_but( P, Excl, Trans ) :-
    findall( A, next( P, A, _ ), PossA ),
    setminus( PossA, Excl, T1 ),
    ( P = n(K,_) -> 
	del_bound_trans( K, T1, Excl, Trans );
	Trans = T1
    ).

del_bound_trans( _, T, [], T ).
del_bound_trans( Net, T, [E|Es], TE ) :- 
    (
    E =.. [=,St^In,C], bound( S^O, St^In, Net ) -> 
	removefrom( T, S^O:=C, TE1 );
    E = update(St^In), bound( S^O, St^In, Net ) -> 
	removefrom( T, S^O:=C, T1 ), 
	removefrom( T1, S^O:==_, T2 ),
	removefrom( T2, send(S^O), TE1 )
    ),
    del_bound_trans( Net, TE1, Es, TE ).
% TODO: filter out complementary events of outputs

% removefrom( A, X, B ): remove element X from a list A
removefrom( [], _, [] ).
removefrom( [X|Tail], X, Tail ) :- !.
removefrom( [Y|Tail], X, [Y|L] ) :- removefrom( Tail, X, L ).

% setminus( A, B, D ): true if D = A \ B
setminus( A, [], A ) :- !.
setminus( A, [B|Bs], D ) :- removefrom( A, B, D1 ), setminus( D1, Bs, D ).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%
% Debugging
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% print all possible runs up to level Max
run( P, Max ) :- run( P, 0, Max ).

run( _, Level, Max ) :- Level > Max, !.
run( P, Level, Max ) :-
    indent( Level ), 
    write( '--> ' ), write( P ), nl,
    findall( [A,P1], next( P, A, P1 ), Sol ), 
    L1 is Level+1, runall( Sol, L1, Max ), !.

% run for all given transition alternatives
runall( [], _, _ ).
runall( [[A,P]|T], Level, Max ) :-
    indent( Level ), write( A ), nl,
    run( P, Level, Max ), runall( T, Level, Max ).

% print N indentations
indent( 0 ).
indent( N ) :- write( '    ' ), N1 is N-1, indent( N1 ).

% print every step of the given action list
trace( _, [] ).
trace( P, [A|As] ) :-
    write(A), nl, 
    next( P, A, P1 ),
    write(P1), nl,
    trace(P1, As).

